# flake8: noqa
# nopycln: file
import cucu.hooks

import cucu.steps.section_steps
import cucu.steps.base_steps

import cucu.steps.browser_steps
import cucu.steps.button_steps
import cucu.steps.command_steps
import cucu.steps.checkbox_steps
import cucu.steps.draggable_steps
import cucu.steps.dropdown_steps
import cucu.steps.filesystem_steps
import cucu.steps.image_steps
import cucu.steps.input_steps
import cucu.steps.file_input_steps
import cucu.steps.link_steps
import cucu.steps.flow_control_steps
import cucu.steps.menuitem_steps
import cucu.steps.platform_steps
import cucu.steps.radio_steps
import cucu.steps.tab_steps
import cucu.steps.table_steps
import cucu.steps.text_steps
import cucu.steps.variable_steps
import cucu.steps.webserver_steps



import logging

import humanize
from selenium.common.exceptions import (
    ElementClickInterceptedException,
    ElementNotInteractableException,
)
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.ui import Select
from tenacity import (
    before_sleep_log,
    retry_if_exception_type,
    retry_if_result,
    stop_after_attempt,
    wait_fixed,
)
from tenacity import retry as retrying

from cucu import fuzzy, helpers, logger, retry, step
from cucu.steps.input_steps import find_input
from cucu.utils import take_saw_element_screenshot

from . import base_steps


def find_dropdown(ctx, name, index=0):
    """
    find a dropdown on screen by fuzzy matching on the name provided and the
    target element:

        * <select>
        * <* role="combobox">
        * <* role="listbox">

    parameters:
      ctx(object): behave context object used to share data between steps
      name(str):   name that identifies the desired dropdown on screen
      index(str):  the index of the dropdown if there are duplicates

    returns:
        the WebElement that matches the provided arguments.
    """
    ctx.check_browser_initialized()

    dropdown = fuzzy.find(
        ctx.browser,
        name,
        [
            "select",
            '*[role="combobox"]',
            '*[role="listbox"]',
        ],
        index=index,
        direction=fuzzy.Direction.LEFT_TO_RIGHT,
    )
    if not dropdown:
        # In case the name is on the top of the dropdown element,
        # the name is after the ting in DOM. Try the other direction.
        dropdown = fuzzy.find(
            ctx.browser,
            name,
            [
                "select",
                '*[role="combobox"]',
                '*[role="listbox"]',
            ],
            index=index,
            direction=fuzzy.Direction.RIGHT_TO_LEFT,
        )

    take_saw_element_screenshot(ctx, "dropdown", name, index, dropdown)

    if dropdown:
        outer_html = dropdown.get_attribute("outerHTML")
        logger.debug(f'looked for dropdown "{name}", and found "{outer_html}"')
    else:
        logger.debug(f'looked for dropdown "{name}" but found none')

    return dropdown


@retrying(
    retry=retry_if_result(lambda result: result is None),
    stop=stop_after_attempt(10),
    wait=wait_fixed(0.1),
    before_sleep=before_sleep_log(logger, logging.DEBUG),
    reraise=True,
    retry_error_callback=lambda retry_state: retry_state.outcome.result(),
)
def find_dropdown_option(ctx, name, index=0):
    """
    find a dropdown option with the provided name. It only considers
    the web element with the name inside the element.

        * <option>
        * <* role="option">

    parameters:
      ctx(object): behave context object used to share data between steps
      name(str):   name that identifies the desired dropdown on screen
      index(str):  the index of the dropdown if there are duplicates

    returns:
        the WebElement that matches the provided arguments.
    """
    ctx.check_browser_initialized()

    option = fuzzy.find(
        ctx.browser,
        name,
        [
            "option",
            '*[role="option"]',
            '*[role="treeitem"]',
            "*[aria-selected]",
        ],
        index=index,
        direction=fuzzy.Direction.LEFT_TO_RIGHT,
        name_within_thing=True,
    )
    if option:
        outer_html = option.get_attribute("outerHTML")
        logger.debug(
            f'looked for dropdown option "{name}", and found "{outer_html}"'
        )
        take_saw_element_screenshot(ctx, "option", name, index, option)
    else:
        logger.debug(f'looked for dropdown option "{name}" but found none')

    return option


def click_dropdown(ctx, dropdown):
    """
    Internal method used to simply click a dropdown element

    Args:
        ctx(object): behave context object used to share data between steps
        dropdown(WebElement): the dropdown element
    """
    ctx.check_browser_initialized()

    if base_steps.is_disabled(dropdown):
        raise RuntimeError("unable to click the button, as it is disabled")

    logger.debug("clicking dropdown")
    try:
        ctx.browser.click(dropdown)
    except ElementClickInterceptedException:
        clickable = dropdown
        while True:
            # In some cases, the dropdown is blocked by the selected item.
            # It finds the ancestors of the dropdown that is clickable and click.
            clickable = clickable.find_element(By.XPATH, "..")
            try:
                ctx.browser.click(clickable)
            except ElementClickInterceptedException:
                continue
            break


@retrying(
    retry=retry_if_exception_type(ElementNotInteractableException),
    stop=stop_after_attempt(10),
    wait=wait_fixed(0.1),
    before_sleep=before_sleep_log(logger, logging.DEBUG),
    reraise=True,
)
def click_dynamic_dropdown_option(ctx, option_element):
    ctx.browser.execute("arguments[0].scrollIntoView();", option_element)
    ctx.browser.click(option_element)


def find_n_select_dropdown_option(ctx, dropdown, option, index=0):
    """
    find and select dropdown option

    parameters:
      ctx(object): behave context object used to share data between steps
      name(str):   name that identifies the desired dropdown on screen
      option(str): name of the option to select
      index(str):  the index of the dropdown if there are duplicates
    """
    ctx.check_browser_initialized()

    dropdown_element = find_dropdown(ctx, dropdown, index)

    if dropdown_element is None:
        prefix = "" if index == 0 else f"{humanize.ordinal(index + 1)} "
        raise RuntimeError(f"unable to find the {prefix}dropdown {dropdown}")

    if base_steps.is_disabled(dropdown_element):
        raise RuntimeError(
            "unable to select from the dropdown, as it is disabled"
        )

    if dropdown_element.tag_name == "select":
        select_element = Select(dropdown_element)
        select_element.select_by_visible_text(option)

    else:
        if dropdown_element.get_attribute("aria-expanded") != "true":
            # open the dropdown
            click_dropdown(ctx, dropdown_element)

        option_element = find_dropdown_option(ctx, option)

        if option_element is None:
            raise RuntimeError(
                f'unable to find option "{option}" in dropdown "{dropdown}"'
            )

        logger.debug("clicking dropdown option")
        ctx.browser.execute("arguments[0].scrollIntoView();", option_element)
        ctx.browser.click(option_element)


def find_n_select_dynamic_dropdown_option(ctx, dropdown, option, index=0):
    """
    find and select dynamic dropdown option

    parameters:
      ctx(object): behave context object used to share data between steps
      name(str):   name that identifies the desired dropdown on screen
      option(str): name of the option to select
      index(str):  the index of the dropdown if there are duplicates
    """
    ctx.check_browser_initialized()

    dropdown_element = find_dropdown(ctx, dropdown, index)

    if dropdown_element is None:
        prefix = "" if index == 0 else f"{humanize.ordinal(index + 1)} "
        raise RuntimeError(f"unable to find the {prefix}dropdown {dropdown}")

    if base_steps.is_disabled(dropdown_element):
        raise RuntimeError(
            "unable to select from the dropdown, as it is disabled"
        )

    if dropdown_element.get_attribute("aria-expanded") != "true":
        logger.debug("open the dropdown")
        click_dropdown(ctx, dropdown_element)

    option_element = find_dropdown_option(ctx, option)

    # Use the search feature to make the option visible so cucu can pick it up
    if option_element is None:
        dropdown_input = find_input(ctx, dropdown, index)
        logger.debug(
            f'option "{option}" is not found, trying to send keys "{option}".'
        )
        dropdown_value = dropdown_input.get_attribute("value")
        if dropdown_value:
            logger.debug(f"clear dropdown value: {dropdown_value}")
            dropdown_input.send_keys(
                Keys.ARROW_RIGHT * len(dropdown_value)
            )  # make sure the cursor is at the end
            dropdown_input.send_keys(Keys.BACKSPACE * len(dropdown_value))
        # After each key stroke there is a request and an update of the option list. To prevent stale element,
        # we send keys one by one here and try to find the option after each key.
        for key in option:
            try:
                dropdown_input = find_input(ctx, dropdown, index)
                logger.debug(f'sending key "{key}"')
                dropdown_input.send_keys(key)
                ctx.browser.wait_for_page_to_load()
                option_element = find_dropdown_option(ctx, option)
                if option_element:
                    break
            except Exception:
                option_element = None

    if option_element is None:
        raise RuntimeError(
            f'unable to find option "{option}" in dropdown "{dropdown}"'
        )

    logger.debug("clicking dropdown option")
    click_dynamic_dropdown_option(ctx, option_element)


def assert_dropdown_option_selected(
    ctx, dropdown, option, index=0, is_selected=True
):
    """
    assert dropdown option is selected

    parameters:
      ctx(object): behave context object used to share data between steps
      name(str):   name that identifies the desired dropdown on screen
      option(str): name of the option to select
      index(str):  the index of the dropdown if there are duplicates
    """
    ctx.check_browser_initialized()

    dropdown_element = find_dropdown(ctx, dropdown, index)
    if dropdown_element is None:
        raise RuntimeError(f'unable to find dropdown "{dropdown}"')

    selected_option = None
    if dropdown_element.tag_name == "select":
        select_element = Select(dropdown_element)
        selected_option = select_element.first_selected_option

        if selected_option is None:
            raise RuntimeError(
                f"unable to find selected option in dropdown {dropdown}"
            )

        selected_name = selected_option.get_attribute("textContent")

        # XXX: we're doing contains because a lot of our existing dropdowns
        #      do not use aria-label/aria-describedby to make them accessible
        #      and easier to find for automation by their name
        if is_selected:
            if selected_name.find(option) == -1:
                raise RuntimeError(f"{option} is not selected")
        else:
            if selected_name.find(option) != -1:
                raise RuntimeError(f"{option} is selected")

    else:
        if dropdown_element.get_attribute("aria-expanded") != "true":
            # open the dropdown to see its options
            click_dropdown(ctx, dropdown_element)

        selected_option = find_dropdown_option(ctx, option)

        if selected_option is None:
            raise RuntimeError(
                f'unable to find option "{option}" in dropdown "{dropdown}"'
            )

        if is_selected:
            if selected_option.get_attribute("aria-selected") != "true":
                raise RuntimeError(f"{option} is not selected")
        else:
            if selected_option.get_attribute("aria-selected") == "true":
                raise RuntimeError(f"{option} is selected")

        # close the dropdown
        click_dropdown(ctx, dropdown_element)


helpers.define_should_see_thing_with_name_steps("dropdown", find_dropdown)
helpers.define_thing_with_name_in_state_steps(
    "dropdown",
    "disabled",
    find_dropdown,
    base_steps.is_disabled,
    with_nth=True,
)
helpers.define_thing_with_name_in_state_steps(
    "dropdown",
    "not disabled",
    find_dropdown,
    base_steps.is_not_disabled,
    with_nth=True,
)
helpers.define_run_steps_if_I_can_see_element_with_name_steps(
    "dropdown", find_dropdown
)
helpers.define_action_on_thing_with_name_steps(
    "dropdown", "click", find_dropdown, click_dropdown, with_nth=True
)
helpers.define_thing_with_name_in_state_steps(
    "dropdown option", "disabled", find_dropdown_option, base_steps.is_disabled
)


@step('I select the option "{option}" from the dropdown "{dropdown}"')
def select_option_from_dropdown(ctx, option, dropdown):
    find_n_select_dropdown_option(ctx, dropdown, option)


@step(
    'I select the option "{option}" from the "{index:nth}" dropdown "{dropdown}"'
)
def select_option_from_nth_dropdown(ctx, option, dropdown, index):
    find_n_select_dropdown_option(ctx, dropdown, option, index)


@step(
    'I wait to select the option "{option}" from the "{index:nth}" dropdown "{dropdown}"'
)
def wait_to_select_option_from_nth_dropdown(ctx, option, dropdown, index):
    retry(find_n_select_dropdown_option)(ctx, dropdown, option, index)


@step('I wait to select the option "{option}" from the dropdown "{dropdown}"')
def wait_to_select_option_from_dropdown(ctx, option, dropdown):
    retry(find_n_select_dropdown_option)(ctx, dropdown, option)


@step('I select the option "{option}" from the dynamic dropdown "{dropdown}"')
def select_option_from_dynamic_dropdown(ctx, option, dropdown):
    find_n_select_dynamic_dropdown_option(ctx, dropdown, option)


@step(
    'I select the option "{option}" from the "{index:nth}" dynamic dropdown "{dropdown}"'
)
def select_option_from_nth_dynamic_dropdown(ctx, option, dropdown, index):
    find_n_select_dynamic_dropdown_option(ctx, dropdown, option, index)


@step(
    'I wait to select the option "{option}" from the "{index:nth}" dynamic dropdown "{dropdown}"'
)
def wait_to_select_option_from_nth_dynamic_dropdown(
    ctx, option, dropdown, index
):
    retry(find_n_select_dynamic_dropdown_option)(ctx, dropdown, option, index)


@step(
    'I wait to select the option "{option}" from the dynamic dropdown "{dropdown}"'
)
def wait_to_select_option_from_dynamic_dropdown(ctx, option, dropdown):
    retry(find_n_select_dynamic_dropdown_option)(ctx, dropdown, option)


@step(
    'I should see the option "{option}" is selected on the dropdown "{dropdown}"'
)
def should_see_option_is_selected_from_dropdown(ctx, option, dropdown):
    assert_dropdown_option_selected(ctx, dropdown, option, is_selected=True)


@step(
    'I should see the option "{option}" is selected on the "{index:nth}" dropdown "{dropdown}"'
)
def should_see_option_is_selected_from_nth_dropdown(
    ctx, option, dropdown, index
):
    assert_dropdown_option_selected(
        ctx, dropdown, option, index, is_selected=True
    )


@step(
    'I wait to see the option "{option}" is selected on the dropdown "{dropdown}"'
)
def wait_to_see_option_is_selected_from_dropdown(ctx, option, dropdown):
    retry(assert_dropdown_option_selected)(
        ctx, dropdown, option, is_selected=True
    )


@step(
    'I should see the option "{option}" is not selected on the "{index:nth}" dropdown "{dropdown}"'
)
def should_see_option_is_not_selected_from_nth_dropdown(
    ctx, option, dropdown, index
):
    assert_dropdown_option_selected(
        ctx, dropdown, option, index, is_selected=False
    )


@step(
    'I should see the option "{option}" is not selected on the dropdown "{dropdown}"'
)
def should_see_option_is_not_selected_from_dropdown(ctx, option, dropdown):
    assert_dropdown_option_selected(ctx, dropdown, option, is_selected=False)


@step(
    'I wait to see the option "{option}" is not selected on the dropdown "{dropdown}"'
)
def wait_to_see_option_is_not_selected_from_dropdown(
    ctx, option, dropdown, is_selected=False
):
    retry(assert_dropdown_option_selected)(ctx, dropdown, option)



# -*- coding: utf-8 -*-
__version__ = "0.1.0"

import sys
from cucu import behave_tweaks
from cucu.browser import selenium_tweaks

# intercept the stdout/stderr so we can do things such as hiding secrets in logs
behave_tweaks.init_step_hooks(sys.stdout, sys.stderr)
selenium_tweaks.init()

from cucu.hooks import (
    init_global_hook_variables,
    init_scenario_hook_variables,
    register_after_all_hook,
    register_before_all_hook,
    register_after_this_scenario_hook,
    register_before_scenario_hook,
    register_after_scenario_hook,
    register_before_step_hook,
    register_after_step_hook,
    register_page_check_hook,
    register_custom_variable_handling,
    register_custom_tags_in_report_handling,
    register_custom_scenario_subheader_in_report_handling,
    register_custom_junit_failure_handler,
    register_before_retry_hook,
)

from cucu.utils import (
    format_gherkin_table,
    run_steps,
    retry,
    StopRetryException,
)

from cucu import helpers
from behave import step



import operator
import re
import sys
import time

import parse
from behave import register_type
from behave.model_describe import ModelPrinter

from cucu import logger, step
from cucu.ansi_parser import remove_ansi
from cucu.config import CONFIG

NTH_REGEX = r"(\d+)(nd|th|rd|st)"


@parse.with_pattern(NTH_REGEX)
def parse_nth(nth):
    matcher = re.match(NTH_REGEX, nth)

    if matcher is None:
        raise Exception(f"nth expression {nth} is invalid")

    number, _ = matcher.groups()
    return int(number) - 1


register_type(nth=parse_nth)


def is_disabled(element):
    """
    internal method to check an element is disabled
    """
    return (
        element.get_attribute("disabled")
        or element.get_attribute("aria-disabled") == "true"
    )


def is_not_disabled(element):
    """
    internal method to check an element is not disabled
    """
    return not is_disabled(element)


@step("I run a step that fails")
def this_step_fails(_):
    raise Exception("failing on purpose")


@step('I sleep for "{value}" seconds')
def sleep(ctx, value):
    time.sleep(int(value))


@step('I echo "{value}"')
def i_echo(ctx, value):
    print(f"{value}\n")


@step("I echo the following")
def i_echo_the_following(ctx):
    if ctx.text is not None:
        print(f"{ctx.text}\n")

    elif ctx.table is not None:
        printer = ModelPrinter(sys.stdout)
        printer.print_table(ctx.table)
        print("")


@step('I log "{message}" at level "{level}"')
def i_log(_, message, level):
    operator.methodcaller(level.lower(), message)(logger)


@step('I log the following at level "{level}"')
def i_log_following(ctx, level):
    operator.methodcaller(level.lower(), ctx.text)(logger)


@step(
    'I strip ansi codes from "{value}" and save to the variable "{variable}"'
)
def strip_ansi_codes_and_save(ctx, value, variable):
    CONFIG[variable] = remove_ansi(value)



import logging
import re
from importlib import metadata

import jellyfish
from lsprotocol.types import (
    TEXT_DOCUMENT_COMPLETION,
    CompletionItem,
    CompletionList,
    CompletionOptions,
    CompletionParams,
)
from pygls.lsp.server import LanguageServer

from cucu import init_global_hook_variables
from cucu.cli.steps import load_cucu_steps

init_global_hook_variables()


def find_completions(step_fragment, steps_cache=None):
    """
    given a step name fragment return all of the possible step name completions
    for that fragment.

    Params:
        step_fragment(string): prefix of the step name to look for such as
                             "I click"

    Returns:
        an Array tuples of step name and locations.
    """
    items = []

    if steps_cache is None:
        steps_cache, _ = load_cucu_steps()

    # first pass try to find steps that start with fragment provided
    for step in steps_cache.keys():
        if step.startswith(step_fragment):
            step_details = steps_cache[step]
            location = step_details["location"]
            location = f"{location['filepath']}:{location['line']}"
            items.append((step, location))

    # if there were 0 steps found then lets at least find some that are close
    # based on some string distance heuristic
    if len(items) == 0:
        for step in steps_cache.keys():
            if jellyfish.jaro_similarity(step_fragment, step) > 0.6:
                step_details = steps_cache[step]
                location = step_details["location"]
                location = f"{location['filepath']}:{location['line']}"
                items.append((step, location))

        def compare_completion_item(completion_item):
            return jellyfish.jaro_similarity(completion_item[0], step_fragment)

        items.sort(key=compare_completion_item, reverse=True)

    return items


def start(port=None):
    version = metadata.version(__package__.split(".")[0])
    server = LanguageServer(name="cucu", version=version)
    steps_cache, _ = load_cucu_steps()
    logging.basicConfig(
        filename="pygls.log", filemode="w", level=logging.DEBUG
    )

    @server.feature(
        TEXT_DOCUMENT_COMPLETION, CompletionOptions(trigger_characters=[","])
    )
    def completions(ls: LanguageServer, params: CompletionParams):
        logging.warn(f"{TEXT_DOCUMENT_COMPLETION}: {params}")

        document_uri = params.text_document.uri
        document = document_lines = ls.workspace.get_document(document_uri)
        document_lines = document.source.split("\n")

        completion_line = document_lines[params.position.line]
        logging.debug(f'completions for "{completion_line}"')

        completion_line = completion_line.lstrip()
        step_line = re.match(
            "(Given|When|Then|And) (.*)", completion_line
        ).groups()[1]

        items = []
        step_completions = find_completions(
            completion_line, steps_cache=steps_cache
        )

        for step_name, step_location in step_completions:
            insert_text = step_name.replace(step_line, "")
            items.append(
                CompletionItem(
                    label=step_name,
                    detail=f"defined in {step_location}",
                    insert_text=insert_text,
                )
            )

        return CompletionList(
            is_incomplete=False,
            items=items,
        )

    if port:
        server.start_tcp("0.0.0.0", int(port))

    else:
        server.start_io()



import os
import re
import subprocess
import time

from cucu import logger, retry, run_steps, step
from cucu.config import CONFIG
from cucu.hooks import register_after_this_scenario_hook


@step('I skip this scenario if the file at "{filepath}" exists')
def skip_scenario_if_file_exists(ctx, filepath):
    if os.path.exists(filepath):
        ctx.scenario.skip(
            reason='skipping scenario since file at "{filepath}" exists'
        )


@step('I expect the following step to fail with "{message}"')
def expect_the_following_step_to_fail(ctx, message):
    try:
        run_steps(ctx, ctx.text)
    except Exception as exception:
        if str(exception).find(message) == -1:
            raise RuntimeError(
                f'expected failure message was "{str(exception)}" not "{message}"'
            )
        return

    raise RuntimeError("previous steps did not fail!")


@step('I should see the previous step took less than "{seconds}" seconds')
def should_see_previous_step_took_less_than(ctx, seconds):
    if ctx.scenario.previous_step_duration > float(seconds):
        raise RuntimeError(
            f"previous step took {ctx.scenario.previous_step_duration}, which is more than {seconds}"
        )


@step('I should see the previous step took more than "{seconds}" seconds')
def should_see_previous_step_took_more_than(ctx, seconds):
    if ctx.scenario.previous_step_duration < float(seconds):
        raise RuntimeError(
            f"previous step took {ctx.scenario.previous_step_duration}, which is less than {seconds}"
        )


@step("I wait to see the following steps succeed")
def wait_to_see_the_following_steps_suceed(ctx):
    retry(run_steps)(ctx, ctx.text)


@step('I wait up to "{seconds}" seconds to see the following steps succeed')
def wait_up_to_seconds_to_see_the_following_steps_suceed(ctx, seconds):
    retry(run_steps, wait_up_to_s=float(seconds))(ctx, ctx.text)


def wait_for_steps_to_fail(ctx, steps, timeout=None):
    def steps_should_fail():
        try:
            run_steps(ctx, steps)
        except:  # noqa: E722
            return

        raise RuntimeError("underlying steps did not fail")

    retry(steps_should_fail, wait_up_to_s=timeout)()


@step("I wait to see the following steps fail")
def wait_to_see_the_following_steps_fail(ctx):
    wait_for_steps_to_fail(ctx, ctx.text)


@step('I wait up to "{seconds}" seconds to see the following steps fail')
def wait_up_to_seconds_to_see_the_following_steps_fail(ctx, seconds):
    wait_for_steps_to_fail(ctx, ctx.text, timeout=float(seconds))


def repeat_steps(ctx, steps, repeat, variable=None):
    for index in range(0, repeat):
        if variable is not None:
            CONFIG[variable] = index + 1

        run_steps(ctx, steps)


@step('I repeat "{repeat}" times the following steps')
def repeat_n_times_the_following_steps(ctx, repeat):
    repeat_steps(ctx, ctx.text, int(repeat))


@step(
    'I repeat "{repeat}" times the following steps with iteration variable "{variable}"'
)
def repeat_n_times_the_following_steps_with_variable(ctx, repeat, variable):
    repeat_steps(ctx, ctx.text, int(repeat), variable=variable)


@step('I run the following steps if the file at "{filepath}" does not exist')
def run_steps_if_file_does_not_exist(ctx, filepath):
    if not os.path.exists(filepath):
        run_steps(ctx, ctx.text)


@step('I run the following steps if the file at "{filepath}" exists')
def run_steps_if_file_exists(ctx, filepath):
    if os.path.exists(filepath):
        run_steps(ctx, ctx.text)


@step('I run the following timed steps as "{name}"')
def run_and_measure_the_following_steps(ctx, name):
    start = time.time()
    run_steps(ctx, ctx.text)
    duration = round(time.time() - start, 3)
    logger.info(f'"{name}" timer took {duration}s')


@step('I start the timer "{name}"')
def start_the_timer(ctx, name):
    ctx.step_timers[name] = time.time()


@step('I stop the timer "{name}"')
def stop_the_timer(ctx, name):
    if name not in ctx.step_timers:
        raise RuntimeError(f'no previously started timer with name "{name}"')
    start = ctx.step_timers[name]
    del ctx.step_timers[name]

    duration = round(time.time() - start, 3)
    logger.info(f'"{name}" timer took {duration}s')


def run_feature(ctx, filepath, results):
    command = f"cucu run {filepath} --results {results}"
    process = subprocess.run(command, shell=True)  # nosec

    return_code = process.returncode
    if return_code != 0:
        raise RuntimeError(
            f'"{command}" exited with {return_code}, see above for details'
        )


@step(
    'I run the feature at "{feature_filepath}" with results at "{results_filepath}"'
)
def run_feature_at(ctx, feature_filepath, results_filepath):
    run_feature(ctx, feature_filepath, results_filepath)


@step(
    'I run the feature at "{feature_filepath}" with results at "{results_filepath}" if the file at "{filepath}" does not exist'
)
def run_feature_if_file_does_not_exist(
    ctx, feature_filepath, results_filepath, filepath
):
    if not os.path.exists(filepath):
        run_feature(ctx, feature_filepath, results_filepath)


@step(
    'I run the feature at "{feature_filepath}" with results at "{results_filepath}" if the file at "{filepath}" exists'
)
def run_feature_if_file_exists(
    ctx, feature_filepath, results_filepath, filepath
):
    if os.path.exists(filepath):
        run_feature(ctx, feature_filepath, results_filepath)


@step("I run the following steps at the end of the current scenario")
def run_the_following_steps_at_end_of_scenario(ctx):
    steps = ctx.text

    def run_final_steps(ctx):
        run_steps(ctx, steps)

    register_after_this_scenario_hook(run_final_steps)


for operation, operator in {
    "is equal to": lambda x, y: x == y,
    "is not equal to": lambda x, y: x != y,
    "contains": lambda x, y: y in x,
    "does not contain": lambda x, y: y not in x,
    "matches": lambda x, y: re.match(y, x),
    "does not match": lambda x, y: not (re.match(y, x)),
}.items():
    # the operator keyword below is used to scope the value of the operator
    # value to the correct value at runtime when creating the various steps
    @step('I run the following steps if "{this}" ' + operation + ' "{that}"')
    def run_steps_if_this_operator_that(ctx, this, that, operator=operator):
        if operator(this, that):
            run_steps(ctx, ctx.text)



import json
import sys
import traceback
from functools import partial
from pathlib import Path

import yaml

from cucu import config, init_scenario_hook_variables, logger
from cucu.config import CONFIG
from cucu.page_checks import init_page_checks
from cucu.utils import (
    TeeStream,
    ellipsize_filename,
    get_iso_timestamp_with_ms,
    parse_iso_timestamp,
    take_screenshot,
)

CONFIG.define(
    "SCENARIO_RESULTS_DIR",
    "the results directory for the currently executing scenario",
    default=None,
)
CONFIG.define(
    "SCENARIO_DOWNLOADS_DIR",
    "the browser downloads directory for the currently executing scenario",
    default=None,
)


init_page_checks()


def check_browser_initialized(ctx):
    """
    check browser session initialized otherwise throw an exception indicating
    such to be used consistently by all steps in this module.
    """

    if ctx.browser is None:
        raise RuntimeError("browser not currently open")


def before_all(ctx):
    CONFIG["__CUCU_CTX"] = ctx
    ctx.check_browser_initialized = partial(check_browser_initialized, ctx)
    ctx.worker_custom_data = {}

    CONFIG.snapshot("before_all")

    for hook in CONFIG["__CUCU_BEFORE_ALL_HOOKS"]:
        hook(ctx)


def after_all(ctx):
    # run the after all hooks
    for hook in CONFIG["__CUCU_AFTER_ALL_HOOKS"]:
        hook(ctx)

    CONFIG.restore(with_pop=True)


def before_feature(ctx, feature):
    if config.CONFIG["CUCU_RESULTS_DIR"] is not None:
        results_dir = Path(config.CONFIG["CUCU_RESULTS_DIR"])
        ctx.feature_dir = results_dir / ellipsize_filename(feature.name)


def after_feature(ctx, feature):
    pass


def before_scenario(ctx, scenario):
    # we want every scenario to start with the exact same reinitialized config
    # values and not bleed values between scenario runs
    CONFIG.restore()

    # we should load any cucurc.yml files in the path to the feature file
    # we are about to run so that the config values set for this feature are
    # correctly loaded.
    CONFIG.load_cucurc_files(ctx.feature.filename)

    init_scenario_hook_variables()

    ctx.scenario = scenario
    ctx.step_index = 0
    ctx.browsers = []
    ctx.browser = None
    ctx.section_step_stack = []

    # reset the step timer dictionary
    ctx.step_timers = {}
    scenario.start_at = get_iso_timestamp_with_ms()

    if config.CONFIG["CUCU_RESULTS_DIR"] is not None:
        ctx.scenario_dir = ctx.feature_dir / ellipsize_filename(scenario.name)
        CONFIG["SCENARIO_RESULTS_DIR"] = ctx.scenario_dir
        ctx.scenario_dir.mkdir(parents=True, exist_ok=True)

        ctx.scenario_downloads_dir = ctx.scenario_dir / "downloads"
        CONFIG["SCENARIO_DOWNLOADS_DIR"] = ctx.scenario_downloads_dir
        ctx.scenario_downloads_dir.mkdir(parents=True, exist_ok=True)

        ctx.scenario_logs_dir = ctx.scenario_dir / "logs"
        CONFIG["SCENARIO_LOGS_DIR"] = ctx.scenario_logs_dir
        ctx.scenario_logs_dir.mkdir(parents=True, exist_ok=True)

        cucu_debug_log_path = ctx.scenario_logs_dir / "cucu.debug.console.log"
        ctx.scenario_debug_log_file = open(
            cucu_debug_log_path, "w", encoding=sys.stdout.encoding
        )
        ctx.scenario_debug_log_tee = TeeStream(ctx.scenario_debug_log_file)

        # redirect stdout, stderr and setup a logger at debug level to fill
        # the scenario cucu.debug.log file which makes it possible to have
        # debug logging for every single scenario run without polluting the
        # console logs at runtime.
        sys.stdout.set_other_stream(ctx.scenario_debug_log_tee)
        sys.stderr.set_other_stream(ctx.scenario_debug_log_tee)
        logger.init_debug_logger(ctx.scenario_debug_log_tee)

        # capture browser logs using TeeStream since each call clears the log
        ctx.browser_log_file = open(
            ctx.scenario_logs_dir / "browser_console.log.txt",
            "w",
            encoding="utf-8",
        )
        ctx.browser_log_tee = TeeStream(ctx.browser_log_file)

    # run before all scenario hooks
    for hook in CONFIG["__CUCU_BEFORE_SCENARIO_HOOKS"]:
        try:
            hook(ctx)
            logger.debug(f"HOOK {hook.__name__}: passed ✅")
        except Exception as e:
            error_message = (
                f"HOOK-ERROR in {hook.__name__}: {e.__class__.__name__}: {e}\n"
            )
            error_message += traceback.format_exc()
            logger.error(error_message)
            ctx.scenario.mark_skipped()
            # Set 'hook_failed' status to 'True' so that the test gets marked
            # as 'errored', even though no steps ran
            ctx.scenario.hook_failed = True


def run_after_scenario_hook(ctx, scenario, hook):
    try:
        hook(ctx)
        logger.debug(f"HOOK {hook.__name__}: passed ✅")
    except Exception as e:
        # For any after scenario hooks,'hook_failed' status will be 'False'
        # but will attach the error message to scenario.
        error_message = (
            f"HOOK-ERROR in {hook.__name__}: {e.__class__.__name__}: {e}\n"
        )
        error_message += traceback.format_exc()
        logger.error(error_message)


def after_scenario(ctx, scenario):
    for timer_name in ctx.step_timers:
        logger.warning(f'timer "{timer_name}" was never stopped/recorded')

    browser_info = {"has_browser": False}

    if len(ctx.browsers) != 0:
        try:
            tab_info = ctx.browser.get_tab_info()
            all_tabs = ctx.browser.get_all_tabs_info()
            browser_info = {
                "has_browser": True,
                "current_tab_index": tab_info["index"],
                "all_tabs": all_tabs,
                "browser_type": ctx.browser.driver.name,
                "session_id": ctx.browser.get_session_id(),
            }
        except Exception as e:
            logger.error(f"Error getting browser info: {e}")
    scenario.browser_info = browser_info

    run_after_scenario_hook(ctx, scenario, download_mht_data)

    # run after all scenario hooks in 'lifo' order.
    for hook in CONFIG["__CUCU_AFTER_SCENARIO_HOOKS"][::-1]:
        run_after_scenario_hook(ctx, scenario, hook)

    # run after this scenario hooks in 'lifo' order.
    for hook in CONFIG["__CUCU_AFTER_THIS_SCENARIO_HOOKS"][::-1]:
        run_after_scenario_hook(ctx, scenario, hook)

    CONFIG["__CUCU_AFTER_THIS_SCENARIO_HOOKS"] = []

    if CONFIG.true("CUCU_KEEP_BROWSER_ALIVE"):
        logger.debug("keeping browser alive between sessions")
    elif len(ctx.browsers) != 0:
        logger.debug("quitting browser between sessions")
        run_after_scenario_hook(ctx, scenario, cleanup_browsers)

    cucu_config_path = ctx.scenario_logs_dir / "cucu.config.yaml.txt"
    with open(cucu_config_path, "w") as config_file:
        config_file.write(CONFIG.to_yaml_without_secrets())

    scenario.cucu_config_json = yaml.safe_load(
        CONFIG.to_yaml_without_secrets()
    )

    scenario.end_at = get_iso_timestamp_with_ms()


def download_mht_data(ctx):
    if not ctx.browsers:
        logger.debug("No browsers - skipping MHT webpage snapshot")
    elif config.CONFIG["CUCU_BROWSER"].lower() != "chrome":
        logger.debug("Browser not Chrome - skipping MHT webpage snapshot")
    else:
        for index, browser in enumerate(ctx.browsers):
            mht_filename = (
                f"browser{index if len(ctx.browsers) > 1 else ''}_snapshot.mht"
            )
            mht_pathname = CONFIG["SCENARIO_LOGS_DIR"] / mht_filename
            logger.debug(f"Saving MHT webpage snapshot: {mht_filename}")
            browser.download_mht(mht_pathname)


def cleanup_browsers(ctx):
    # close the browser unless someone has set the keep browser alive
    # environment variable which allows tests to reuse the same browser
    # session

    for browser in ctx.browsers:
        browser.quit()

    ctx.browser_log_file.close()

    ctx.browsers = []


def before_step(ctx, step):
    step.start_at = get_iso_timestamp_with_ms()

    sys.stdout.captured()
    sys.stderr.captured()

    # Reset the debug log buffer for this step
    if hasattr(ctx, "scenario_debug_log_tee"):
        ctx.scenario_debug_log_tee.clear()

    ctx.current_step = step
    step.is_substep = getattr(step, "is_substep", False)
    step.has_substeps = getattr(step, "has_substeps", False)
    ctx.section_level = None
    step.seq = ctx.step_index + 1
    step.parent_seq = (
        ctx.section_step_stack[-1].seq if ctx.section_step_stack else 0
    )

    CONFIG["__STEP_SCREENSHOT_COUNT"] = 0

    # run before all step hooks
    for hook in CONFIG["__CUCU_BEFORE_STEP_HOOKS"]:
        hook(ctx)


def after_step(ctx, step):
    step.stdout = sys.stdout.captured()
    step.stderr = sys.stderr.captured()

    # Capture debug output from the TeeStream for this step
    if hasattr(ctx, "scenario_debug_log_tee"):
        step.debug_output = ctx.scenario_debug_log_tee.getvalue().splitlines()
    else:
        step.debug_output = []

    step.end_at = get_iso_timestamp_with_ms()

    # calculate duration from ISO timestamps
    ctx.scenario.previous_step_duration = (
        parse_iso_timestamp(step.end_at) - parse_iso_timestamp(step.start_at)
    ).total_seconds()

    # when set this means we're running in parallel mode using --workers and
    # we want to see progress reported using simply dots
    if CONFIG["__CUCU_PARENT_STDOUT"]:
        CONFIG["__CUCU_PARENT_STDOUT"].write(".")
        CONFIG["__CUCU_PARENT_STDOUT"].flush()

    # we only take screenshots of steps where there's a browser currently open
    # and this step has no substeps as in the reporting the substeps that
    # may actually do something on the browser take their own screenshots
    if ctx.browser is not None and ctx.current_step.has_substeps is False:
        take_screenshot(ctx, step.name, label=f"After {step.name}")

        tab_info = ctx.browser.get_tab_info()
        total_tabs = tab_info["tab_count"]
        current_tab = tab_info["index"] + 1
        title = tab_info["title"]
        url = tab_info["url"]
        log_message = (
            f"\ntab({current_tab} of {total_tabs}): {title}\nurl: {url}\n"
        )
        logger.debug(log_message)

        # Add tab info to step.stdout so it shows up in the HTML report
        step.stdout.extend(
            [f"tab({current_tab} of {total_tabs}): {title}", f"url: {url}"]
        )

    # if the step has substeps from using `run_steps` then we already moved
    # the step index in the run_steps method and shouldn't do it here
    if not step.has_substeps:
        ctx.step_index += 1
        CONFIG["__STEP_SCREENSHOT_COUNT"] = 0

    if CONFIG.bool("CUCU_DEBUG_ON_FAILURE") and step.status == "failed":
        ctx._runner.stop_capture()
        import pdb

        pdb.post_mortem(step.exc_traceback)

    CONFIG["__CUCU_BEFORE_THIS_SCENARIO_HOOKS"] = []

    # run after all step hooks
    for hook in CONFIG["__CUCU_AFTER_STEP_HOOKS"]:
        hook(ctx)

    # Capture browser logs and info for this step
    step.browser_logs = []
    browser_info = {"has_browser": False}
    if ctx.browser:
        for log in ctx.browser.get_log():
            log_entry = json.dumps(log)
            step.browser_logs.append(log)
            ctx.browser_log_tee.write(f"{log_entry}\n")

        tab_info = ctx.browser.get_tab_info()

        browser_info = {
            "tab_count": tab_info["tab_count"],
            "tab_number": tab_info["index"] + 1,
            "tab_title": tab_info["title"],
            "tab_url": tab_info["url"],
            "browser_type": ctx.browser.driver.name,
        }

    step.browser_info = browser_info



import pkgutil
from enum import Enum

from cucu import logger
from cucu.browser.frames import search_in_all_frames


def load_jquery_lib():
    """
    load jquery library
    """
    jquery_lib = pkgutil.get_data(
        "cucu", "external/jquery/jquery-3.5.1.min.js"
    )
    return jquery_lib.decode("utf8")


def load_fuzzy_lib():
    """
    load the fuzzy javascript library
    """
    return pkgutil.get_data("cucu", "fuzzy/fuzzy.js").decode("utf8")


def init(browser):
    """
    initializes the fuzzy matching javascript library within the currently open
    browsers execution engine

    parameters:
        browser - ...
    """
    script = "return typeof cucu !== 'undefined' && typeof cucu.fuzzy_find === 'function';"
    cucu_injected = browser.execute(script)
    if cucu_injected:
        # cucu fuzzy find already exists
        return

    logger.debug("inject cucu fuzzy find library to the browser")
    jqCucu_script = "window.jqCucu = jQuery.noConflict(true);"
    browser.execute(load_jquery_lib() + jqCucu_script + load_fuzzy_lib())


class Direction(Enum):
    """
    simple Direction enum
    """

    LEFT_TO_RIGHT = 1
    RIGHT_TO_LEFT = 2


def find(
    browser,
    name,
    things,
    index=0,
    direction=Direction.LEFT_TO_RIGHT,
    name_within_thing=False,
):
    """
    find an element by applying the fuzzy finding rules when given the name
    that identifies the element on screen and a list of possible `things` that
    are CSS expression fragments like so:

        tag_name[attribute expression]

    That identify the kind of element you're trying to find, such as a button,
    input[type='button'], etc.

    parameters:
      browser           - the cucu.browser.Browser object
      name              - name that identifies the element you are trying to find
      things            - array of CSS fragments that specify the kind of elements you
                          want to match on
      index             - which of the many matches to return
      direction         - the text to element direction to apply fuzzy in. Default we
                          apply right to left but for checkboxes or certain languages
                          this direction can be used to find things by prioritizing
                          matching from "left to right"
      name_within_thing - to determine if the name has to be within the web element

    returns:
        the WebElement that matches the provided arguments.
    """
    browser.switch_to_default_frame()

    # always need to protect names in which double quotes are used as below
    # we pass arguments to the fuzzy_find javascript function wrapped in double
    # quotes
    name = name.replace('"', '\\"')
    name_within_thing = "true" if name_within_thing else "false"

    args = [
        f'"{name}"',
        str(things),
        str(index),
        str(direction.value),
        name_within_thing,
        "true",
    ]

    def execute_fuzzy_find():
        init(browser)
        script = f"return cucu.fuzzy_find({','.join(args)});"
        return browser.execute(script)

    fuzzy_return = search_in_all_frames(browser, execute_fuzzy_find)
    if fuzzy_return is None:
        logger.debug("Fuzzy found no element.")
        return None
    logger.debug(
        "Fuzzy found element by search term {}".format(fuzzy_return[1])
    )
    return fuzzy_return[0]



import logging

from behave import use_step_matcher

from cucu import step

use_step_matcher("re")  # use regex to match section heading patterns


@step("(#{1,4})\\s*(.*)")
def section_step(ctx, section_level, section_text):
    """
    A section heading step that organizes scenarios into logical sections.

    Usage:
        * # Main section
        * ## Subsection
        * ### Sub-subsection
        * #### Deep subsection

    The number of # characters determines the heading level (1-4).
    This step is a no-op but provides structure in the HTML report.
    """
    step = ctx.current_step
    step.section_level = len(section_level)
    step.parent_seq = 0

    while len(ctx.section_step_stack):
        latest_section = ctx.section_step_stack[-1]
        if latest_section.section_level < step.section_level:
            step.parent_seq = latest_section.seq
            break
        ctx.section_step_stack.pop()
        logging.debug(
            f"Section: exited '{latest_section.name}' (level {latest_section.section_level})"
        )

    ctx.section_step_stack.append(ctx.current_step)
    logging.debug(
        f"Section: entering '{step.name}' (level {step.section_level})"
    )


use_step_matcher("parse")  # set this back to cucu's default matcher parser



# -*- coding: utf-8 -*-
import os
import shutil
import signal
import sqlite3
import sys
import time
import xml.etree.ElementTree as ET
from collections import Counter
from importlib.metadata import version
from pathlib import Path
from threading import Timer

import click
import coverage
import psutil
from behave.model_core import FileLocation
from behave.runner_util import parse_features
from click import ClickException
from mpire import WorkerPool
from tabulate import tabulate

from cucu import (
    fuzzy,
    init_global_hook_variables,
    language_server,
    logger,
    register_after_all_hook,
    reporter,
)
from cucu.cli import thread_dumper
from cucu.cli.run import behave, behave_init, create_run
from cucu.cli.steps import print_human_readable_steps, print_json_steps
from cucu.config import CONFIG
from cucu.db import (
    consolidate_database_files,
    finish_worker_record,
)
from cucu.lint import linter
from cucu.utils import generate_short_id

# will start coverage tracking once COVERAGE_PROCESS_START is set
coverage.process_startup()


@click.group()
@click.version_option(version("cucu"), message="%(version)s")
def main():
    """
    cucu e2e testing framework
    """
    pass


@main.command()
@click.argument(
    "filepath", default="features", type=click.Path(path_type=Path)
)
@click.option(
    "-b",
    "--browser",
    default=os.environ.get("CUCU_BROWSER") or "chrome",
    help="browser name to use default: chrome",
)
@click.option(
    "-c",
    "--color-output/--no-color-output",
    default=True,
    help="produce output with colors or not",
)
@click.option(
    "--dry-run/--no-dry-run",
    default=False,
    help="invokes output formatters without running the steps",
)
@click.option(
    "-e",
    "--env",
    default=[],
    multiple=True,
    help="set environment variable which can be referenced with",
)
@click.option(
    "-g",
    "--generate-report/--no-generate-report",
    default=False,
    help="automatically generate a report at the end of the test run",
)
@click.option(
    "-x",
    "--fail-fast/--no-fail-fast",
    default=False,
    help="stop running tests on the first failure",
)
@click.option(
    "-h",
    "--headless/--no-headless",
    default=True,
    help="controls if the browser is run in headless mode",
)
@click.option("-n", "--name", help="used to specify the exact scenario to run")
@click.option(
    "-d",
    "--debug-on-failure/--no-debug-on-failure",
    default=False,
    help="on failure drop into the debug shell",
)
@click.option(
    "-j",
    "--junit",
    default=None,
    help="specify the output directory for JUnit XML files, default is "
    "the same location as --results",
    type=click.Path(path_type=Path),
)
@click.option(
    "--junit-with-stacktrace",
    is_flag=True,
    default=False,
    help="when set to true the JUnit XML output will contain the stacktrace",
)
@click.option(
    "-l",
    "--logging-level",
    default="INFO",
    help="set logging level to one of debug, warn or info (default)",
)
@click.option(
    "--show-skips",
    default=False,
    is_flag=True,
    help="when set skips are shown",
)
@click.option(
    "--periodic-thread-dumper",
    default=None,
    help="sets the interval in minutes of when to run the periodic thread dumper",
)
@click.option(
    "-p",
    "--preserve-results/--no-preserve-results",
    default=False,
    help="when set we will not remove any existing results directory",
)
@click.option(
    "--record-env-vars",
    default=False,
    is_flag=True,
    help="when set will record shell environment variables to debug file: run_details.json",
)
@click.option(
    "--report",
    default="report",
    help="the location to put the test report when --generate-report is used",
    type=click.Path(path_type=Path),
)
@click.option(
    "-r",
    "--results",
    default="results",
    help="the results directory used by cucu",
    type=click.Path(path_type=Path),
)
@click.option(
    "--runtime-timeout",
    default=None,
    type=int,
    help="the runtime timeout in seconds after which the current run will terminate any running tests and exit",
)
@click.option(
    "--feature-timeout",
    default=1800,
    help="When run tests in parallel, the maximum amount of time (seconds) a feature can run",
)
@click.option(
    "--secrets",
    default=None,
    help="coma separated list of variable names that we should hide"
    " their value all of the output produced by cucu",
)
@click.option(
    "-t",
    "--tags",
    default=[],
    multiple=True,
    help="Only execute features or scenarios with tags matching "
    "expression provided. example: --tags @dev, --tags ~@dev",
)
@click.option(
    "-w",
    "--workers",
    default=None,
    help="Specifies the number of workers to use to run tests in parallel",
    type=int,
)
@click.option(
    "--verbose/--no-verbose",
    default=False,
    help="runs with verbose logging and shows additional stacktrace",
)
@click.option(
    "-s",
    "--selenium-remote-url",
    default=None,
    help="the HTTP url for a selenium hub setup to run the browser tests on",
)
def run(
    filepath,
    browser,
    color_output,
    dry_run,
    env,
    generate_report,
    fail_fast,
    headless,
    name,
    debug_on_failure,
    junit,
    junit_with_stacktrace,
    logging_level,
    periodic_thread_dumper,
    preserve_results,
    record_env_vars,
    report,
    results,
    runtime_timeout,
    feature_timeout,
    secrets,
    show_skips,
    tags,
    selenium_remote_url,
    workers,
    verbose,
):
    """
    run a set of feature files

    Note: All the os.environ variables are set to be available in the child processes
    """
    init_global_hook_variables()
    dumper = None

    if os.environ.get("CUCU") == "true":
        # when cucu is already running it means that we're running inside
        # another cucu process and therefore we should make sure the results
        # directory isn't the default one and throw an exception otherwise
        if results == Path("results"):
            raise Exception(
                "running within cucu but --results was not used, "
                "this would lead to some very difficult to debug "
                "failures as this process would clobber the "
                "parent results directory"
            )

    # set for testing cucu itself but basically allows you to know when cucu
    # is running itself as part of internal testing
    os.environ["CUCU"] = "true"

    os.environ["CUCU_LOGGING_LEVEL"] = logging_level.upper()
    logger.init_logging(logging_level.upper())

    if not preserve_results:
        if results.exists():
            shutil.rmtree(results)

    results.mkdir(parents=True, exist_ok=True)

    if selenium_remote_url is not None:
        os.environ["CUCU_SELENIUM_REMOTE_URL"] = selenium_remote_url

    if periodic_thread_dumper is not None:
        interval_min = float(periodic_thread_dumper)
        dumper = thread_dumper.start(interval_min)

    # need to set this before initializing any browsers below
    os.environ["CUCU_BROWSER"] = browser.lower()

    if junit is None:
        junit = results

    if show_skips:
        os.environ["CUCU_SHOW_SKIPS"] = "true"

    if junit_with_stacktrace:
        os.environ["CUCU_JUNIT_WITH_STACKTRACE"] = "true"

    if record_env_vars:
        os.environ["CUCU_RECORD_ENV_VARS"] = "true"

    cucu_run_id_seed = f"{time.perf_counter()}_{os.getpid()}"
    os.environ["CUCU_RUN_ID"] = CONFIG["CUCU_RUN_ID"] = generate_short_id(
        cucu_run_id_seed
    )
    worker_id_seed = f"{CONFIG['CUCU_RUN_ID']}_{os.getpid()}"
    os.environ["WORKER_PARENT_ID"] = CONFIG["WORKER_RUN_ID"] = (
        generate_short_id(worker_id_seed)
    )

    os.environ["CUCU_FILEPATH"] = CONFIG["CUCU_FILEPATH"] = str(filepath)

    create_run(results, filepath)

    try:
        if workers is None or workers == 1:
            logger.debug(
                f"Starting cucu_run {CONFIG['CUCU_RUN_ID']} with single worker"
            )
            if runtime_timeout:
                logger.debug("setting up runtime timeout timer")

                def runtime_exit():
                    logger.error("runtime timeout reached, aborting run")
                    CONFIG["__CUCU_CTX"]._runner.aborted = True
                    os.kill(os.getpid(), signal.SIGINT)

                timer = Timer(runtime_timeout, runtime_exit)
                timer.start()

                def cancel_timer(_):
                    logger.debug("cancelled runtime timeout timer")
                    timer.cancel()

                register_after_all_hook(cancel_timer)

            exit_code = behave(
                filepath,
                color_output,
                dry_run,
                env,
                fail_fast,
                headless,
                name,
                debug_on_failure,
                junit,
                results,
                secrets,
                show_skips,
                tags,
                verbose,
                skip_init_global_hook_variables=True,
            )

            if exit_code != 0:
                raise ClickException("test run failed, see above for details")

        else:
            logger.debug(
                f"Starting cucu_run {CONFIG['CUCU_RUN_ID']} with multiple workers: {workers}"
            )
            if filepath.is_dir():
                feature_filepaths = list(filepath.rglob("*.feature"))
            else:
                feature_filepaths = [filepath]

            if sys.platform == "darwin":
                logger.info(
                    "MAC OS detected, using 'forkserver' start method since 'fork' is unstable"
                )
                start_method = "forkserver"
            else:
                start_method = "fork"

            with WorkerPool(
                n_jobs=int(workers), start_method=start_method
            ) as pool:
                # Each feature file is applied to the pool as an async task.
                # It then polls the async result of each task. It the result
                # is ready, it removes the result from the list of results that
                # need to be checked again until all the results are checked.
                # If the timer is triggered, it stops the while loop and
                # logs all the unfinished features.
                # The pool is terminated automatically when it exits the
                # context.
                timer = None
                timeout_reached = False
                if runtime_timeout:
                    logger.debug("setting up runtime timeout timer")

                    def runtime_exit():
                        nonlocal timeout_reached
                        logger.error("runtime timeout reached, aborting run")
                        timeout_reached = True

                    timer = Timer(runtime_timeout, runtime_exit)
                    timer.start()

                def kill_workers():
                    for worker in pool._workers:
                        try:
                            worker_proc = psutil.Process(worker.pid)
                            for child in worker_proc.children():
                                child.kill()

                            worker_proc.kill()
                        except psutil.NoSuchProcess:
                            pass

                def handle_kill_signal(signum, frame):
                    signal.signal(
                        signum, signal.SIG_IGN
                    )  # ignore additional signals
                    logger.warning(
                        f"received signal {signum}, sending kill signal to workers"
                    )
                    kill_workers()
                    if timer:
                        timer.cancel()

                    os.kill(os.getpid(), signal.SIGINT)

                # This is for local runs where you want to cancel the run with a ctrl+c or SIGTERM
                signal.signal(signal.SIGINT, handle_kill_signal)
                signal.signal(signal.SIGTERM, handle_kill_signal)

                async_results = {}
                for feature_filepath in feature_filepaths:
                    async_results[feature_filepath] = pool.apply_async(
                        behave,
                        [
                            feature_filepath,
                            color_output,
                            dry_run,
                            env,
                            fail_fast,
                            headless,
                            name,
                            debug_on_failure,
                            junit,
                            results,
                            secrets,
                            show_skips,
                            tags,
                            verbose,
                        ],
                        {
                            "redirect_output": True,
                        },
                        task_timeout=float(feature_timeout),
                    )
                    logger.info(f"scheduled feature file {feature_filepath}")

                # poll while we have running tasks until the overall time limit
                task_failed = {}
                while not timeout_reached:
                    remaining = {}
                    for feature, result in async_results.items():
                        if timeout_reached:
                            break

                        if result.ready():
                            try:
                                # wait 0.1s max for interprocess communication
                                exit_code = result.get(0.1)
                                if exit_code != 0:
                                    task_failed[feature] = result
                            except TimeoutError as err:
                                if f"timeout={feature_timeout}" in str(err):
                                    print(f"{err}")
                                    task_failed[feature] = result
                                # ignore timeout errors from interprocess communication slowness
                            except Exception:
                                logger.exception(
                                    f"an exception is raised during feature {feature}"
                                )
                                task_failed[feature] = result
                        else:
                            remaining[feature] = result

                    async_results = remaining

                    if len(remaining) == 0:
                        if timer:
                            # we're done so cancel any outstanding overall time limit
                            timer.cancel()
                        break

                    time.sleep(1)

                if timeout_reached:
                    logger.warning(
                        "Timeout reached, send kill signal to workers"
                    )
                    kill_workers()

                task_failed.update(async_results)

                if task_failed:
                    failing_features = [str(x) for x in task_failed.keys()]
                    logger.error(
                        f"Failing Features:\n{'\n'.join(failing_features)}"
                    )
                    raise RuntimeError(
                        "there are failures, see above for details"
                    )
    finally:
        if dumper is not None:
            dumper.stop()

        if results.exists():
            finish_worker_record(worker_run_id=CONFIG.get("WORKER_PARENT_ID"))
            consolidate_database_files(results)

        if generate_report:
            _generate_report(
                results_dir=results,
                report_folder=report,
                junit_folder=junit,
            )


def _generate_report(
    results_dir: Path,
    report_folder: Path,
    junit_folder: Path | None = None,
    combine: bool = False,
):
    if report_folder.exists():
        shutil.rmtree(report_folder)

    report_folder.mkdir(parents=True, exist_ok=True)

    if results_dir.exists():
        consolidate_database_files(results_dir, combine)

    report_location = reporter.generate(results_dir, report_folder)
    print(f"HTML test report at {report_location}")

    if junit_folder:
        for junit_file in junit_folder.rglob("*.xml"):
            junit = ET.parse(junit_file)
            test_suite = junit.getroot()
            ts_folder = test_suite.get("foldername")
            for test_case in test_suite.iter("testcase"):
                report_path = os.path.join(
                    report_folder,
                    ts_folder,
                    test_case.get("foldername"),
                    "index.html",
                )
                test_case.set("report_path", report_path)
            junit.write(junit_file, encoding="utf-8", xml_declaration=False)


@main.command()
@click.argument(
    "results_dir", default="results", type=click.Path(path_type=Path)
)
@click.option(
    "-l",
    "--logging-level",
    default="INFO",
    help="set logging level to one of debug, warn or info (default)",
)
@click.option(
    "--show-skips",
    default=False,
    is_flag=True,
    help="when set skips are shown",
)
@click.option(
    "-o",
    "--output",
    default="report",
    type=click.Path(path_type=Path),
)
@click.option(
    "-j",
    "--junit",
    default=None,
    help="specify the output directory for JUnit XML files, default is "
    "the same location as --results",
    type=click.Path(path_type=Path),
)
@click.option(
    "--combine",
    default=False,
    is_flag=True,
    help="combine multiple cucu_runs into a single report",
)
def report(
    results_dir: Path,
    logging_level,
    show_skips,
    output: Path,
    junit: Path,
    combine: bool,
):
    """
    generate a test report from a results directory
    """
    init_global_hook_variables()

    os.environ["CUCU_LOGGING_LEVEL"] = logging_level.upper()
    logger.init_logging(logging_level.upper())

    if show_skips:
        os.environ["CUCU_SHOW_SKIPS"] = "true"

    run_db_path = results_dir / "run.db"
    if run_db_path.exists():
        # query cucu_run to get the original filepath used during the run
        with sqlite3.connect(run_db_path) as conn:
            cursor = conn.cursor()
            cursor.execute(
                "SELECT filepath FROM cucu_run ORDER BY start_at DESC LIMIT 1"
            )
            row = cursor.fetchone()
            if row:
                filepath = row[0]
                # initialize any underlying custom step code things
                behave_init(filepath)

    _generate_report(
        results_dir=results_dir,
        report_folder=output,
        junit_folder=junit,
        combine=combine,
    )


@main.command()
@click.argument(
    "filepath", default="features", type=click.Path(path_type=Path)
)
@click.option(
    "-f",
    "--format",
    default="human",
    help="output format to use, available: human, json."
    "default: human. PRO TIP: `brew install fzf` and then "
    "`cucu steps | fzf` and easily find the step you need.",
)
def steps(filepath, format):
    """
    print available cucu steps
    """
    init_global_hook_variables()

    if format == "human":
        print_human_readable_steps(filepath=filepath)

    elif format == "json":
        print_json_steps(filepath=filepath)

    else:
        raise RuntimeError(f'unsupported format "{format}"')


@main.command()
@click.argument("filepath", type=click.Path(path_type=Path), nargs=-1)
@click.option(
    "--fix/--no-fix", default=False, help="fix lint violations, default: False"
)
@click.option(
    "-l",
    "--logging-level",
    default="INFO",
    help="set logging level to one of debug, warn or info (default)",
)
def lint(filepath, fix, logging_level):
    """
    lint feature files
    """
    os.environ["CUCU_LOGGING_LEVEL"] = logging_level.upper()
    logger.init_logging(logging_level.upper())

    init_global_hook_variables()

    logger.init_logging("INFO")
    filepaths = list(filepath)

    if filepaths == []:
        filepaths = ["features"]

    violations_found = 0
    violations_fixed = 0

    for filepath in filepaths:
        # initialize any underlying custom step code things
        behave_init(filepath)

        all_violations = linter.lint(filepath)

        for violations in all_violations:
            if fix:
                violations = linter.fix(violations)

            if violations:
                for violation in violations:
                    violations_found += 1

                    if violation["type"] == "steps_error":
                        print(violation["message"])
                        print(
                            "failure loading some steps, see above for details"
                        )
                        print("")
                        continue

                    location = violation["location"]
                    _type = violation["type"][0].upper()
                    message = violation["message"]
                    suffix = ""

                    if fix:
                        if violation["fixed"]:
                            suffix = " ✓"
                            violations_fixed += 1
                        else:
                            suffix = " ✗ (must be fixed manually)"

                    filepath = location["filepath"]
                    line_number = location["line"] + 1
                    print(
                        f"{filepath}:{line_number}: {_type} {message}{suffix}"
                    )

    if violations_found != 0:
        if violations_found == violations_fixed:
            print("\nlinting errors found and fixed, see above for details")

        else:
            raise ClickException(
                "linting errors found, but not fixed, see above for details"
            )


@main.command()
@click.option(
    "-l",
    "--logging-level",
    default="INFO",
    help="set logging level to one of debug, warn or info (default)",
)
@click.option(
    "-p",
    "--port",
    default=None,
    help="when the port is set the lsp will run in TCP mode and not STDIO mode",
)
def lsp(logging_level, port):
    """
    start the cucu language server
    """
    os.environ["CUCU_LOGGING_LEVEL"] = logging_level.upper()
    logger.init_logging(logging_level.upper())

    language_server.start(port=port)


@main.command()
@click.argument(
    "filepath", default="features", type=click.Path(path_type=Path)
)
def vars(filepath):
    """
    print built-in cucu variables
    """
    init_global_hook_variables()

    # loading the steps make it so the code that registers config variables
    # elsewhere get to execute
    behave_init(filepath)

    variables = []
    variables.append(["Name", "Description", "Default"])

    variables.extend(
        [
            [name, definition["description"], definition["default"]]
            for name, definition in CONFIG.defined_variables.items()
        ]
    )

    print(tabulate(variables, tablefmt="fancy_grid"))


@main.command()
@click.argument("repo_dir", default="", type=click.Path(path_type=Path))
@click.option(
    "-l",
    "--logging-level",
    default="INFO",
    help="set logging level to one of debug, warn or info (default)",
)
def init(repo_dir, logging_level):
    """
    initialize cucu in the current directory

    Copies the `init_data` directory to the current directory
    """
    os.environ["CUCU_LOGGING_LEVEL"] = logging_level.upper()
    logger.init_logging(logging_level.upper())

    init_data_dir = Path(__file__).parent.parent / "init_data"

    logger.debug(f"cucu init: copy example directory from {init_data_dir=}")

    features_dir = repo_dir / "features"
    if features_dir.exists():
        answer = input("Overwrite existing files? [y/N]:")
        if answer.lower() != "y":
            print("Aborted!")
            return

    shutil.copytree(init_data_dir, repo_dir, dirs_exist_ok=True)
    print("You can now start writing your tests")


@main.command()
@click.option(
    "-b",
    "--browser",
    default="chrome",
    help="when specified the browser will be opened with the fuzzy "
    "js library preloaded.",
)
@click.option(
    "-u",
    "--url",
    default="https://www.google.com",
    help="URL to open the browser at for debugging",
)
@click.option(
    "--detach",
    default=False,
    help="when set to detach the browser will continue to run and "
    "the cucu process will exit",
)
@click.option(
    "-l",
    "--logging-level",
    default="INFO",
    help="set logging level to one of debug, warn or info (default)",
)
def debug(browser, url, detach, logging_level):
    """
    debug cucu library
    """
    os.environ["CUCU_LOGGING_LEVEL"] = logging_level.upper()
    logger.init_logging(logging_level.upper())

    fuzzy_js = fuzzy.load_jquery_lib() + fuzzy.load_fuzzy_lib()
    # XXX: need to make this more generic once we make the underlying
    #      browser framework swappable.
    from cucu.browser.selenium import Selenium

    selenium = Selenium()
    selenium.open(browser, detach=detach)
    selenium.navigate(url)
    selenium.execute(fuzzy_js)

    if not detach:
        while True:
            # detect when there are changes to the cucu javascript library
            # and reload it in the currently running browser.
            time.sleep(5)


@main.command()
@click.option(
    "-l",
    "--logging-level",
    default="INFO",
    help="set logging level to one of debug, warn or info (default)",
)
@click.argument(
    "filepath", default="features", type=click.Path(path_type=Path)
)
def tags(filepath, logging_level):
    """
    print a table of tags and affected scenario counts
    """
    init_global_hook_variables()
    os.environ["CUCU_LOGGING_LEVEL"] = logging_level.upper()
    logger.init_logging(logging_level.upper())

    if filepath.is_file():
        feature_files = [filepath]
    else:
        feature_files = list(filepath.rglob("*.feature"))

    if not filepath.exists() or not feature_files:
        raise ClickException("No feature files found.")

    file_locations = [FileLocation(f.absolute()) for f in feature_files]
    features = parse_features(file_locations)
    tag_scenarios = Counter()

    for feature in features:
        for scenario in feature.scenarios:
            affecting_tags = set(feature.tags + scenario.tags)
            tag_scenarios.update(affecting_tags)

    if not tag_scenarios:
        print("No tags found in feature files.")
        return

    table_data = [["Tag", "Scenarios Affected"]] + [
        [tag_name, str(count)]
        for tag_name, count in sorted(
            tag_scenarios.items(), key=lambda x: x[0].lower()
        )
    ]

    print(tabulate(table_data, headers="firstrow", tablefmt="fancy_grid"))


if __name__ == "__main__":
    main()
